/* $Id: crisv32-hcd.c,v 1.31 2007/01/25 11:07:36 hp Exp $
 *
 * ETRAX FS IOP USB 1.1 Host Controller Driver
 *
 * Copyright (C) 2004, 2005, 2007  Axis Communications AB
 *
 * Author: Lars Viklund <larsv@axis.com>
 * Modified by: Andr Spnberg <andres@axis.com>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/spinlock.h>
#include <linux/usb.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/arch/dma.h>
#include "hcd.h"
#include "hub.h"
#include "iop-usb.h"

#define VERSION			"1.10"	     
#define COPYRIGHT		"(c) 2004, 2005, 2006	Axis Communications AB"

extern int iop_fw_load_mpu(unsigned char *);
extern int iop_fw_load_spu(const unsigned char *, unsigned int);

MODULE_DESCRIPTION("ETRAX FS IOP USB 1.1 Host Controller Driver");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Lars Viklund <larsv@axis.com>");

static int crisv32_hcd_init(void);
static void crisv32_hcd_exit(void);

module_init(crisv32_hcd_init);
module_exit(crisv32_hcd_exit);

/* module parameters */

static unsigned int dma_map = 0;
module_param(dma_map, uint, S_IRUGO);
MODULE_PARM_DESC(dma_map, "DMA mapping to use for USB: 0/1");

static unsigned int ports = (1<<0);
module_param(ports, uint, S_IRUGO);
MODULE_PARM_DESC(ports, "Bitmask indicating USB ports to use");

static int int_phy = 0;
module_param(int_phy, int, S_IRUGO);
MODULE_PARM_DESC(int_phy, "Use internal USB PHY for port: 0/1/none");

static char *mpu_fw = NULL;
module_param(mpu_fw, charp, S_IRUGO);
MODULE_PARM_DESC(mpu_fw, "MPU firmware file name");

static char *spu_fw = NULL;
module_param(spu_fw, charp, S_IRUGO);
MODULE_PARM_DESC(spu_fw, "SPU firmware file name");

/* macros for debug output */

#define hcd_dbg(hcd, fmt, args...) \
	dev_dbg(hcd->self.controller, fmt, ## args)
#define hcd_err(hcd, fmt, args...) \
	dev_err(hcd->self.controller, fmt, ## args)
#define hcd_info(hcd, fmt, args...) \
	dev_info(hcd->self.controller, fmt, ## args)
#define hcd_warn(hcd, fmt, args...) \
	dev_warn(hcd->self.controller, fmt, ## args)

#define str_dir(pipe) \
	(usb_pipeout(pipe) ? "out" : "in")
#define str_type(pipe) \
	({								\
		char *s = "?";						\
		switch (usb_pipetype(pipe)) {				\
		case PIPE_ISOCHRONOUS:	s = "iso";  break;		\
		case PIPE_INTERRUPT:	s = "intr"; break;		\
		case PIPE_CONTROL:	s = "ctrl"; break;		\
		case PIPE_BULK:		s = "bulk"; break;		\
		};							\
		s;							\
	})

/* HCD operations */
static irqreturn_t crisv32_hcd_irq(struct usb_hcd *);
static int crisv32_hcd_reset(struct usb_hcd *);
static int crisv32_hcd_start(struct usb_hcd *);
static void crisv32_hcd_stop(struct usb_hcd *);
#ifdef CONFIG_PM
static int crisv32_hcd_suspend(struct device *, u32, u32);
static int crisv32_hcd_resume(struct device *, u32);
#endif /* CONFIG_PM */
static int crisv32_hcd_get_frame(struct usb_hcd *);
static int crisv32_hcd_urb_enqueue(
	struct usb_hcd *, struct usb_host_endpoint *ep, struct urb *, gfp_t);
static int crisv32_hcd_urb_dequeue(struct usb_hcd *, struct urb *);
static void crisv32_hcd_endpoint_disable(
	struct usb_hcd *, struct usb_host_endpoint *ep);
static int crisv32_hcd_hub_status_data(struct usb_hcd *, char *);
static int crisv32_hcd_hub_control(
	struct usb_hcd *, u16, u16, u16, char *, u16);
#ifdef CONFIG_PM
static int crisv32_hcd_hub_suspend(struct usb_hcd *);
static int crisv32_hcd_hub_resume(struct usb_hcd *);
#endif /* CONFIG_PM */
#ifdef CONFIG_USB_OTG
static int crisv32_hcd_start_port_reset(struct usb_hcd *, unsigned);
#endif /* CONFIG_USB_OTG */
static int crisv32_hcd_probe(struct device *);
static int crisv32_hcd_remove(struct device *);

const char hc_name[] = "crisv32_hc";
const char product_desc[] = "ETRAX FS IOP USB Host Controller";

/* HCD description */
struct crisv32_hcd {
	spinlock_t		lock;
	kmem_cache_t	       *descr_cache;
	struct list_head	done_list;
	u8			iop_init_done;
	u8			num_ports;
};

/* HCD-private transfer descriptor */
struct crisv32_hcd_td {
	struct iop_usb_td	td;
	struct urb	       *urb;
	struct usb_hcd	       *hcd;
	u8			last;
	u8			index;
};

/* HCD-private endpoint descriptor */
struct crisv32_hcd_ed {
	struct iop_usb_ed	ed;
	struct usb_host_endpoint *ep;
	atomic_t		lock_count;
};

/* HCD-private per-urb state */
struct urb_priv {
	struct urb	       *urb;
	struct list_head	done_list;
	struct crisv32_hcd_ed  *ed;
	struct crisv32_hcd_td  *td;
	struct iop_usb_td      *completed;
	int			status;
	int			state;
};

/* states for the urb_priv struct */
#define CRISv32_URB_IDLE	0
#define CRISv32_URB_ENQUEUED	1
#define CRISv32_URB_DONE	2

/* the device */
static struct platform_device *crisv32_hc_platform_device;

/* device driver interface */
static struct device_driver crisv32_hc_device_driver = {
	.name =			(char *) hc_name,
	.bus =			&platform_bus_type,

	.probe =		crisv32_hcd_probe,
	.remove =		crisv32_hcd_remove,

#ifdef CONFIG_PM
	.suspend =		crisv32_hcd_suspend,
	.resume =		crisv32_hcd_resume,
#endif /* CONFIG_PM */
};

/* host controller driver interface */
static const struct hc_driver crisv32_hc_driver = 
{
	.description =		hc_name,
	.product_desc =		product_desc,
	.hcd_priv_size =	sizeof(struct crisv32_hcd),

	/* irq handler */
	.irq =			crisv32_hcd_irq,

	.flags =		HCD_USB11,

	/* called to init HCD and root hub */
	.reset =		crisv32_hcd_reset,
	.start =		crisv32_hcd_start,	

	/* cleanly make HCD stop writing memory and doing I/O */
	.stop =			crisv32_hcd_stop,

	/* return current frame number */
	.get_frame_number =	crisv32_hcd_get_frame,

	/* manage i/o requests, device state */
	.urb_enqueue =		crisv32_hcd_urb_enqueue,
	.urb_dequeue =		crisv32_hcd_urb_dequeue,

	/* hw synch, freeing endpoint resources that urb_dequeue can't */
	.endpoint_disable =	crisv32_hcd_endpoint_disable,

	/* root hub support */
	.hub_status_data =	crisv32_hcd_hub_status_data,
	.hub_control =		crisv32_hcd_hub_control,
#ifdef CONFIG_PM
	.hub_suspend =		crisv32_hcd_hub_suspend,
	.hub_resume =		crisv32_hcd_hub_resume,
#endif /* CONFIG_PM */
#ifdef	CONFIG_USB_OTG
	.start_port_reset =	crisv32_hcd_start_port_reset,
#endif /* CONFIG_USB_OTG */
};

/* size of descriptors in slab cache */
#ifndef MAX
#define MAX(x, y)		((x) > (y) ? (x) : (y))
#endif
#define DESCR_SIZE \
	MAX(sizeof(struct crisv32_hcd_ed), sizeof(struct crisv32_hcd_td))

/* root hub status */
static const struct usb_hub_status rh_hub_status = 
{
	.wHubStatus =		0,
	.wHubChange =		0,
};

/* root hub descriptor */
static const u8 rh_hub_descr[] =
{
	0x09,			/* bDescLength	       */
	0x29,			/* bDescriptorType     */
	IOP_USB_ROOT_HUB_PORTS, /* bNbrPorts	       */
	0x00,			/* wHubCharacteristics */
	0x00,		 
	0x01,			/* bPwrOn2pwrGood      */
	0x00,			/* bHubContrCurrent    */
	0x00,			/* DeviceRemovable     */
	0xff			/* PortPwrCtrlMask     */
};

/* interrupts used by the HCD */
#define INTR_MASK ( \
	(1 << IOP_USB_IRQ_TRANS_DONE) | \
	(1 << IOP_USB_IRQ_BAD_SCHED)  | \
	(1 << IOP_USB_IRQ_INTERNAL))

/* forward declarations of local functions */
static int map_status(int);
static void td_list_free(struct usb_hcd *, struct crisv32_hcd_td *);
static struct crisv32_hcd_td *td_alloc(
	struct usb_hcd *, struct urb *, int, int);
static int td_list_create(
	struct usb_hcd *, struct urb *, struct urb_priv *, int);
static int urb_priv_create(
	struct usb_hcd *, struct urb *, struct crisv32_hcd_ed *, int);
static void urb_priv_free(struct usb_hcd *, struct urb *);
static void finish_urb(struct usb_hcd *, struct urb *);
static void td_complete(struct iop_usb_td *, int, int);
static struct crisv32_hcd_ed *get_ed(
	struct usb_hcd *, struct usb_host_endpoint *, struct usb_device *, 
	unsigned int, int, int);

/*
 * conversion between pointers to a hcd and the corresponding
 * crisv32_hcd 
 */

static __inline__ struct crisv32_hcd *hcd_to_crisv32_hcd(struct usb_hcd *hcd)
{
	return (struct crisv32_hcd *) hcd->hcd_priv;
}

static __inline__ struct usb_hcd *crisv32_hcd_to_hcd(struct crisv32_hcd *hcd)
{
	return container_of((void *) hcd, struct usb_hcd, hcd_priv);
}

/* check if a list contains two or more elements */
static __inline__ int list_two_or_more(const struct list_head *head)
{
	return head->next != head->prev;
}

/* check HCD 'last' flag for a transfer descriptor */
static __inline__ int iop_usb_td_last(struct iop_usb_td *td) 
{  
	return container_of(td, struct crisv32_hcd_td, td)->last;
}

/* map from iop-usb status codes to Linux kernel errors */
static int map_status(int status)
{
	switch (status) {
	case IOP_USB_STAT_OK:
		return 0;
	case IOP_USB_STAT_STALL:
		return -EPIPE;
	case IOP_USB_STAT_E_BITSTUFF:
		return -EPROTO;
	case IOP_USB_STAT_E_PID_CHK:
	case IOP_USB_STAT_E_CRC:
		return -EILSEQ;
	case IOP_USB_STAT_E_BUS_TO:
		return -ETIMEDOUT;
	case IOP_USB_STAT_BABBLE:
	case IOP_USB_STAT_ORUN:
		return -EOVERFLOW;
	case IOP_USB_STAT_URUN:
		return -ENOSR;
	case IOP_USB_STAT_FLUSHED:
		return 0;
	default:	   
		return -EINVAL;
	}
}

/* check if specified port is in use */
static __inline__ int port_in_use(unsigned int port)
{
	return ports & (1 << port);
}

/* numer of ports in use */
static __inline__ unsigned int num_ports(void)
{
	unsigned int i, num = 0;
	for (i = 0; i < IOP_USB_ROOT_HUB_PORTS; i++)
		if (port_in_use(i))
			num++;
	return num;
}

/* map hub port number to the port number used internally by the HC */
static __inline__ unsigned int map_port(unsigned int port)
{
	unsigned int i, num = 0;
	for (i = 0; i < IOP_USB_ROOT_HUB_PORTS; i++)
		if (port_in_use(i))
			if (++num == port)
				return i;
	return -1;
}

/* free a list of transfer descriptors */
static void td_list_free(struct usb_hcd *hcd, struct crisv32_hcd_td *list)
{
	kmem_cache_t *descr_cache = hcd_to_crisv32_hcd(hcd)->descr_cache;

	if (!list)
		return;
	while (1) {
		int last = list->last;
		struct crisv32_hcd_td *next;
		next = (struct crisv32_hcd_td *) list->td.u.next;
		kmem_cache_free(descr_cache, list);
		if (last)
			break;
		list = next;
	}
}

/* allocate a transfer descriptor */
static struct crisv32_hcd_td *td_alloc(
	struct usb_hcd *hcd, 
	struct urb *urb, 
	int index, 
	int mem_flags)
{
	kmem_cache_t *descr_cache = hcd_to_crisv32_hcd(hcd)->descr_cache;
	struct crisv32_hcd_td *td;

	td = kmem_cache_alloc(descr_cache, mem_flags);
	if (!td)
		return NULL;
	td->urb	  = urb;
	td->hcd	  = hcd;
	td->last  = 0;
	td->index = index;
	return td;
}

/* allocate and initialize transfer descriptor lists for an URB */
static int td_list_create(
	struct usb_hcd *hcd,
	struct urb *urb, 
	struct urb_priv *urb_priv, 
	int mem_flags)
{
	int is_in = !!usb_pipein(urb->pipe);
	int type = usb_pipetype(urb->pipe);
	int maxsize = usb_maxpacket(urb->dev, urb->pipe, !is_in);
	int buf_len = urb->transfer_buffer_length;
	void *buf = buf_len > 0 ? urb->transfer_buffer : NULL;
	int i, zeropacket = 0, retval = 0;
	struct crisv32_hcd_td *td;

#define ALLOC_NEXT(t, i) \
	do {								      \
		struct crisv32_hcd_td *new_td;				      \
		new_td = td_alloc(hcd, urb, (i), mem_flags);		      \
		if (!new_td) {						      \
			retval = -ENOMEM;				      \
			goto out;					      \
		}							      \
		(t)->td.u.next = (struct iop_usb_td *) new_td;		      \
		(t) = new_td;						      \
	} while (0)

	td = td_alloc(hcd, urb, 0, mem_flags);
	if (!td)
		return -ENOMEM;
	urb_priv->td = td;

	switch (type) {
	case PIPE_CONTROL:
		/* setup stage */
		iop_usb_trans_init(
			&td->td,
			IOP_USB_SETUP,
			8,
			urb->setup_packet,
			8, 
			0);
		/* data stage */
		if (buf_len > 0 && buf) {
			ALLOC_NEXT(td, 0);
			iop_usb_trans_init(
				&td->td,
				is_in ? IOP_USB_IN : IOP_USB_OUT,
				maxsize,
				buf,
				buf_len,
				0);
		}
		/* status stage */
		ALLOC_NEXT(td, 0);
		iop_usb_trans_init(
			&td->td,
			!is_in ? IOP_USB_IN : IOP_USB_OUT, 
			0,
			NULL,
			0,
			0);
		break;
	case PIPE_ISOCHRONOUS:
		/* the UHCI driver skips zero-sized ISO transfer
		 * buffers, not sure if this is the right thing... */
		for (i = 0; ; i++) {
			iop_usb_trans_init(
				&td->td,
				is_in ? IOP_USB_IN : IOP_USB_OUT,
				maxsize,
				buf + urb->iso_frame_desc[i].offset,
				urb->iso_frame_desc[i].length,
				0);
			if (i >= urb->number_of_packets - 1)
				break;
			ALLOC_NEXT(td, i + 1);
		}
		break;
	case PIPE_BULK:
		zeropacket = urb->transfer_flags & URB_ZERO_PACKET;
		/*FALLTHROUGH*/
	case PIPE_INTERRUPT:
		iop_usb_trans_init(
			&td->td,
			is_in ? IOP_USB_IN : IOP_USB_OUT,
			maxsize,
			buf,
			buf_len, 
			zeropacket);
		break;
	}

 out:
	td->last++;
	if (retval != 0)
		td_list_free(hcd, td);
	return retval;
}

/* allocate and initialize URB private data */
static int urb_priv_create(
	struct usb_hcd *hcd, 
	struct urb *urb, 
	struct crisv32_hcd_ed *ed,
	int mem_flags)
{
	struct urb_priv *urb_priv;
	int retval;

	urb_priv = kmalloc(sizeof *urb_priv, mem_flags);
	if (!urb_priv)
		return -ENOMEM;
	urb_priv->urb = urb;
	urb_priv->ed = ed;
	retval = td_list_create(hcd, urb, urb_priv, mem_flags);
	if (retval != 0) {
		kfree(urb_priv);
		return retval;
	}
	urb_priv->completed = NULL;
	urb_priv->status = 0;
	urb_priv->state = CRISv32_URB_IDLE;
	urb->hcpriv = urb_priv;
	return 0;
}

/* free URB private data */
static void urb_priv_free(struct usb_hcd *hcd, struct urb *urb)
{
	struct urb_priv *urb_priv = urb->hcpriv;

	BUG_ON(!urb_priv);
	td_list_free(hcd, urb_priv->td);
	kfree(urb_priv);
	urb->hcpriv = NULL;
}

/* return URB to driver */
static void finish_urb(
	struct usb_hcd *hcd, 
	struct urb *urb)
{
	struct urb_priv *urb_priv = urb->hcpriv;
	struct usb_host_endpoint *ep = urb_priv->ed->ep;
	unsigned long flags;

	spin_lock_irqsave(&urb->lock, flags);
	if (urb->status == -EINPROGRESS)
		urb->status = urb_priv->status;
	urb_priv_free(hcd, urb);
	spin_unlock_irqrestore(&urb->lock, flags);
	
	/* bandwidth should not be released if it has been propagated
	 * to subsequent URBs for the same endpoint */
	if (urb->bandwidth > 0 && list_two_or_more(&ep->urb_list))
		urb->bandwidth = 0;

	usb_hcd_giveback_urb(hcd, urb);

	/* disable processing of periodic traffic if there is no more
	 * traffic pending */
	if (hcd_to_bus(hcd)->bandwidth_isoc_reqs == 0)
		iop_usb_stop_periodic(IOP_USB_ISOC);
	if (hcd_to_bus(hcd)->bandwidth_int_reqs == 0)
		iop_usb_stop_periodic(IOP_USB_INTR);
}

/* callback invoked by iop_usb_trans_done_irq() for each data
 * descriptor that has been completely processed */
static void td_complete(struct iop_usb_td *td, int status, int length)
{
	struct crisv32_hcd_td *hcd_td = (struct crisv32_hcd_td *) td;
	struct urb *urb = hcd_td->urb;
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd_td->hcd);
	struct urb_priv *urb_priv = urb->hcpriv;

	status = map_status(status);
	urb->actual_length += length;

	/* update URB status if needed */
	switch (usb_pipetype(urb->pipe)) {
	case PIPE_ISOCHRONOUS:
		if (status != 0)
			urb->error_count++;
		urb->iso_frame_desc[hcd_td->index].actual_length += length;
		urb->iso_frame_desc[hcd_td->index].status = status;
		break;
	case PIPE_INTERRUPT:
	case PIPE_CONTROL:
	case PIPE_BULK:
		if (urb_priv->status == 0 && status != 0)
			urb_priv->status = status;
	}

	urb_priv->completed = td;
	if (!hcd_td->last) 
		return;

	/* URB has completed */
	urb_priv->state = CRISv32_URB_DONE;
	if (usb_pipetype(urb->pipe) != PIPE_ISOCHRONOUS && 
	    (urb->transfer_flags & URB_SHORT_NOT_OK) &&
	    urb_priv->status == 0 &&
	    urb->actual_length < urb->transfer_buffer_length)
		urb_priv->status = -EREMOTEIO;
	td->u.next = NULL;
	list_add_tail(&urb_priv->done_list, &crisv32_hcd->done_list);
}

/* get a descriptor for an endpoint, allocate new if not already existing */
static struct crisv32_hcd_ed *get_ed(
	struct usb_hcd *hcd, 
	struct usb_host_endpoint *ep,
	struct usb_device *udev,
	unsigned int pipe,
	int interval,
	int mem_flags)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	kmem_cache_t *descr_cache = crisv32_hcd->descr_cache;
	struct crisv32_hcd_ed *ed = ep->hcpriv;

	if (!ed) {
		int epnum = usb_pipeendpoint(pipe);
		int is_in = !!usb_pipein(pipe);
		int type = usb_pipetype(pipe);
		int daddr = usb_pipedevice(pipe);
		int speed = udev->speed;
		int isoc = type == PIPE_ISOCHRONOUS;
		int maxsize = usb_maxpacket(udev, pipe, !is_in);
		long maxtime = usb_calc_bus_time(speed, is_in, isoc, maxsize);
		unsigned long flags;

		BUG_ON(maxtime < 0);

		/* allocate and initialize a new endpoint descriptor */
		ed = kmem_cache_alloc(descr_cache, mem_flags);
		if (!ed)
			return NULL;
		iop_usb_endpoint_init(
			&ed->ed, type, daddr, epnum, is_in, speed, 
			maxsize, interval, maxtime);
		atomic_set(&ed->lock_count, 0);
		spin_lock_irqsave(&crisv32_hcd->lock, flags);
		iop_usb_endpoint_add(&ed->ed);
		spin_unlock_irqrestore(&crisv32_hcd->lock, flags);
		ed->ep = ep;
		ep->hcpriv = ed;	
		hcd_dbg(hcd, "alloc EP %p: %d %d %s %s\n", 
			ed, daddr, epnum, str_dir(pipe), str_type(pipe));
	}
	return ed;
}

/* interrupt handler */
static irqreturn_t crisv32_hcd_irq(struct usb_hcd *hcd)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	struct urb_priv *urb_priv, *tmp;
	struct list_head done_list;
	int irqs = iop_usb_active_irqs();

	if (irqs & (1 << IOP_USB_IRQ_TRANS_DONE)) {
		unsigned long flags;
		irqs &= ~(1 << IOP_USB_IRQ_TRANS_DONE);
		spin_lock_irqsave(&crisv32_hcd->lock, flags);
		iop_usb_trans_done_irq(td_complete);
		list_add(&done_list, &crisv32_hcd->done_list);
		list_del(&crisv32_hcd->done_list);
		INIT_LIST_HEAD(&crisv32_hcd->done_list);
		spin_unlock_irqrestore(&crisv32_hcd->lock, flags);
		list_for_each_entry_safe(urb_priv, tmp, &done_list, done_list)
			finish_urb(hcd, urb_priv->urb);
	}
	if (irqs & (1 << IOP_USB_IRQ_BAD_SCHED)) {
		irqs &= ~(1 << IOP_USB_IRQ_BAD_SCHED);
		hcd_err(hcd, "Too much periodic traffic committed.\n");
	}
	if (irqs & (1 << IOP_USB_IRQ_INTERNAL)) {
		irqs &= ~(1 << IOP_USB_IRQ_INTERNAL);
		hcd_err(hcd, "Internal error in HC, resetting.\n");
		crisv32_hcd_reset(hcd);
	}
	if (irqs) {
		hcd_err(hcd, "Unexpected interrupt.\n");
	}
	return IRQ_HANDLED;
}

/* reset host controller */
static int crisv32_hcd_reset(struct usb_hcd *hcd)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	int retval;

	spin_lock(&crisv32_hcd->lock);
	retval = iop_usb_reset();
	spin_unlock(&crisv32_hcd->lock);
	if (retval != IOP_USB_OK)
		return -ETIMEDOUT;
	return 0;
}

/* start host controller */
static int crisv32_hcd_start(struct usb_hcd *hcd)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	//struct usb_device *rh_dev;
	int retval = 0;

	spin_lock(&crisv32_hcd->lock);
	retval = iop_usb_start(INTR_MASK);
	spin_unlock(&crisv32_hcd->lock);

	switch (retval) {
	case IOP_USB_OK:
		break;
	case IOP_USB_ERR_TIMEOUT:
		retval = -ETIMEDOUT;
		goto out;
	case IOP_USB_ERR_BAD_FW:
		hcd_warn(hcd, "Bad IOP firmware.\n");
		/*FALLTHROUGH*/
	default:
		retval = -EINVAL;
		goto out;
	}		
	hcd->state = HC_STATE_RUNNING;
	
	if (hcd->self.root_hub)
		return 0;

	return 0;
	
 out:
	hcd->state = HC_STATE_HALT;
	crisv32_hcd_reset(hcd);
	return retval;
}

#ifdef	CONFIG_PM

static int crisv32_hcd_suspend(struct usb_hcd *hcd, u32 state, u32 level)
{
	return 0; /* no-op for now */
}

static int crisv32_hcd_resume(struct usb_hcd *hcd, u32 level)
{
	return 0; /* no-op for now */
}

#endif /* CONFIG_PM */

/* stop host controller */
static void crisv32_hcd_stop(struct usb_hcd *hcd)
{
	if (hcd->irq >= 0)
		disable_irq(hcd->irq);
	crisv32_hcd_reset(hcd);
}

/* return the current frame number */
static int crisv32_hcd_get_frame(struct usb_hcd *hcd)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	int frame;

	spin_lock(&crisv32_hcd->lock);
	frame = iop_usb_get_frame();
	spin_unlock(&crisv32_hcd->lock);
	return frame;
}

/* queue an URB with the host controller */
static int crisv32_hcd_urb_enqueue(
	struct usb_hcd *hcd, 
	struct usb_host_endpoint *ep,
	struct urb *urb, 
	gfp_t mem_flags)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	struct crisv32_hcd_ed *ed;
	struct urb_priv *urb_priv;
	int type = usb_pipetype(urb->pipe);
	int bustime = -1;
	int retval;
	unsigned long flags;

	ed = get_ed(hcd, ep, urb->dev, urb->pipe, urb->interval, mem_flags);
	if (!ed)
		return -ENOMEM;

	retval = urb_priv_create(hcd, urb, ed, mem_flags);
	if (retval != 0)
		return retval;
	urb_priv = urb->hcpriv;

	/* do not submit if the HCD is dead */
	spin_lock(&crisv32_hcd->lock);
	if (!HC_IS_RUNNING(hcd->state)) {
		retval = -ENODEV;
		urb_priv_free(hcd, urb);
		goto out;
	}

	/* check available bandwidth for periodic traffic */
	if (type == PIPE_ISOCHRONOUS || type == PIPE_INTERRUPT) {
		if (list_two_or_more(&ep->urb_list)) {
			/* propagate previously allocated bandwidth */
			struct urb *first_urb = list_entry(
				ep->urb_list.next, struct urb, urb_list);
			urb->bandwidth = first_urb->bandwidth;
		} else {
			/* first URB for EP, allocate bandwidth */
			bustime = usb_check_bandwidth(urb->dev, urb);
			if (bustime < 0) {
				retval = bustime;
				urb_priv_free(hcd, urb);
				goto out;
			}
		}
	}

	/* handle unlink during submit */
	spin_lock(&urb->lock);
	local_irq_save(flags);
	if (urb->status != -EINPROGRESS) {
		local_irq_restore(flags);
		spin_unlock(&urb->lock);
		spin_unlock(&crisv32_hcd->lock);
		finish_urb(hcd, urb);
		return 0;
	}
	
	/* XXX handle start frame for ISOC traffic, not supported in
	 * IOP firmware yet */
	urb->error_count = 0;

	/* insert the new data into the DMA list(s) */
	retval = iop_usb_trans_enqueue(&ed->ed, &urb_priv->td->td);
	if (retval == IOP_USB_OK) {
		retval = 0;
		urb_priv->state = CRISv32_URB_ENQUEUED;
	} else
		retval = -EPIPE;
	local_irq_restore(flags);
	
	/* claim bandwidth for periodic traffic */
	if (retval == 0 && bustime >= 0) 
		usb_claim_bandwidth(
			urb->dev, urb, bustime, type == PIPE_ISOCHRONOUS);
	spin_unlock(&urb->lock);

 out:
	spin_unlock(&crisv32_hcd->lock);
	return retval;
}

/* remove an URB from the host controller queues */
static int crisv32_hcd_urb_dequeue(struct usb_hcd *hcd, struct urb *urb)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	struct crisv32_hcd_ed *ed;
	struct urb_priv *urb_priv;
	struct iop_usb_td *first, *last;
	int retval, dequeued = 0;
	unsigned long flags;

	/* get the endpoint descriptor */
	spin_lock_irqsave(&crisv32_hcd->lock, flags);
	urb_priv = urb->hcpriv;
	if (!urb_priv)
		goto out;
	ed = urb_priv->ed;
	if (!ed) {
		hcd_warn(hcd, "trying to dequeue from non-existing endpoint\n");
		goto out;
	}
	spin_unlock_irqrestore(&crisv32_hcd->lock, flags);

	atomic_inc(&ed->lock_count);
	iop_usb_trans_dequeue_lock(&ed->ed);

	spin_lock_irqsave(&crisv32_hcd->lock, flags);
	urb_priv = urb->hcpriv;
	if (urb_priv && urb_priv->state == CRISv32_URB_ENQUEUED) {
		if (urb_priv->completed) {
			if (iop_usb_td_last(urb_priv->completed))
				goto urb_completed;
			first = urb_priv->completed->u.next;
		} else
			first = &urb_priv->td->td;
		for (last = first;
		     !iop_usb_td_last(last);
		     last = phys_to_virt((unsigned long) last->u.next))
			;
		retval = iop_usb_trans_dequeue(&ed->ed, first, last);
		if (retval == IOP_USB_OK)
			dequeued++;
	}
 urb_completed:
	spin_unlock_irqrestore(&crisv32_hcd->lock, flags);
	if (atomic_dec_and_test(&ed->lock_count))
		iop_usb_trans_dequeue_unlock(&ed->ed);
	if (dequeued)
		finish_urb(hcd, urb);
	return 0;

 out:
	spin_unlock_irqrestore(&crisv32_hcd->lock, flags);
	return 0;
}

/* free resources associated with an endpoint */
static void crisv32_hcd_endpoint_disable(
	struct usb_hcd *hcd, 
	struct usb_host_endpoint *ep)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	struct crisv32_hcd_ed *ed = ep->hcpriv;
	unsigned long flags;

	if (!ed)
		return;
	spin_lock_irqsave(&crisv32_hcd->lock, flags);
	iop_usb_endpoint_release(&ed->ed);
	ep->hcpriv = NULL;
	spin_unlock_irqrestore(&crisv32_hcd->lock, flags);
	iop_usb_endpoint_release_wait(&ed->ed);
	kmem_cache_free(crisv32_hcd->descr_cache, ed);
}

/* construct port status change bitmap for the root hub */
static int crisv32_hcd_hub_status_data(struct usb_hcd *hcd, char *buf)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	u16 wPortChange;
	unsigned int i;

	/*
	 * corresponds to hub status change EP (USB 2.0 spec section 11.13.4)
	 * return bitmap indicating ports with status change
	 */
	*buf = 0;
	spin_lock(&crisv32_hcd->lock);
	for (i = 1; i <= crisv32_hcd->num_ports; i++) {
		int retval = iop_usb_port_stat(
			map_port(i), NULL, &wPortChange);
		if (retval != IOP_USB_OK)
			continue;
		if (wPortChange)
			*buf |= (1 << i);
	}
	spin_unlock(&crisv32_hcd->lock);
	return *buf == 0 ? 0 : 1;
}

/* handle a request for the root hub */
static int crisv32_hcd_hub_control(
	struct usb_hcd *hcd, 
	u16 typeReq, 
	u16 wValue, 
	u16 wIndex,
	char *buf, 
	u16 wLength)
{
	struct crisv32_hcd *crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	u16 wPortStatus, wPortChange;
	size_t len;
	int retval = 0;

	switch (typeReq) {
	case GetHubDescriptor:
		len = min_t(unsigned int, sizeof rh_hub_descr, wLength);
		memcpy(buf, rh_hub_descr, len);
		buf[2] = crisv32_hcd->num_ports;
		break;
	case GetHubStatus:
		len = min_t(unsigned int, sizeof rh_hub_status, wLength);
		memcpy(buf, &rh_hub_status, len);
		break;
	case GetPortStatus:
		if (!wIndex || wIndex > crisv32_hcd->num_ports)
			goto error;
		spin_lock(&crisv32_hcd->lock);
		retval = iop_usb_port_stat(
			map_port(wIndex), &wPortStatus, &wPortChange);
		spin_unlock(&crisv32_hcd->lock);
		if (retval != IOP_USB_OK)
			goto error;
		*(u16 *) buf = cpu_to_le16(wPortStatus);
		*(u16 *) (buf + 2) = cpu_to_le16(wPortChange);
		break;
	case SetHubFeature:
	case ClearHubFeature:
		switch (wValue) {
		case C_HUB_OVER_CURRENT:
		case C_HUB_LOCAL_POWER:
		printk(KERN_INFO "USB: non impl. hub feat: 0x%x\n",wValue);

			/* not implemented */
			break;
		default:
			goto error;
		}
		break;
	case SetPortFeature:
		if (!wIndex || wIndex > crisv32_hcd->num_ports)
			goto error;
		spin_lock(&crisv32_hcd->lock);
		retval = iop_usb_set_port_feat(map_port(wIndex), wValue);
		spin_unlock(&crisv32_hcd->lock);
		if (retval != IOP_USB_OK)
			goto error;
		break;
	case ClearPortFeature:
		if (!wIndex || wIndex > crisv32_hcd->num_ports)
			goto error;
		spin_lock(&crisv32_hcd->lock);
		retval = iop_usb_clr_port_feat(map_port(wIndex), wValue);
		spin_unlock(&crisv32_hcd->lock);
		if (retval != IOP_USB_OK)
			goto error;
		break;
	default:
	error:
		retval = -EPIPE;
	}
	return retval;
}

#ifdef	CONFIG_PM

static int crisv32_hcd_hub_suspend(struct usb_hcd *hcd)
{
	return 0; /* no-op for now */
}

static int crisv32_hcd_hub_resume(struct usb_hcd *hcd)
{
	return 0; /* no-op for now */
}

#endif /* CONFIG_PM */

#ifdef	CONFIG_USB_OTG

static int crisv32_hcd_start_port_reset(struct usb_hcd *hcd, unsigned port)
{
	return 0; /* no-op for now */
}

#endif /* CONFIG_USB_OTG */

/* initialize the IO-processor */
static int __init_or_module crisv32_usb_init_iop(void)
{
	int retval;
	static char fw_name[16];

	/* intialize multiplexing of DMA channels and pins, start IOP clock */
	retval = iop_usb_init(dma_map, ports, int_phy);
	switch (retval) {
	case IOP_USB_OK:
		break;
	case IOP_USB_ERR_BUSY:
		return -EBUSY;
	default:
		return -EINVAL;
	}

	/* load firmware */
	if (!mpu_fw) {
		mpu_fw = fw_name;
		snprintf(mpu_fw, sizeof(fw_name), "usb11-%d.mpu.bin", dma_map);
	}
	retval = iop_fw_load_mpu(mpu_fw);
	if (retval != 0)
		return retval;
	if (!spu_fw) {
		spu_fw = fw_name;
		snprintf(spu_fw, sizeof(fw_name), "usb11-%d.spu.bin", dma_map);
	}
	retval = iop_fw_load_spu(spu_fw, iop_usb_spu_nr());
	if (retval != 0)
		return retval;

	/* configure IOP */
	iop_usb_config();

	return 0;
}

/* initialize the host controller and driver  */
static int __init_or_module crisv32_hcd_probe(struct device *dev)
{
	struct usb_hcd *hcd;
	struct crisv32_hcd *crisv32_hcd;
	int retval;

	hcd = usb_create_hcd(&crisv32_hc_driver, dev, dev->bus_id);
	if (!hcd)
		return -ENOMEM;

	hcd->irq = -1; /* not yet allocated */

	crisv32_hcd = hcd_to_crisv32_hcd(hcd);
	spin_lock_init(&crisv32_hcd->lock);
	INIT_LIST_HEAD(&crisv32_hcd->done_list);
	crisv32_hcd->num_ports = num_ports();
	crisv32_hcd->iop_init_done = 0;
	// Set in usb_create_hcd() but altered here
	dev_set_drvdata(dev, crisv32_hcd);

	crisv32_hcd->descr_cache = 
		kmem_cache_create("iop-usb", DESCR_SIZE, 0, 
				  SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!crisv32_hcd->descr_cache) {
		retval = -ENOMEM;
		goto out;
	}

	retval = crisv32_usb_init_iop();
	if (retval != 0) {
		hcd_err(hcd, "IOP initialization failed\n");
		goto out;
	}
	crisv32_hcd->iop_init_done++;


	retval=usb_add_hcd(hcd,IOP_USB_INTR_VECT,0);

	hcd_info(hcd, "%s\n", hcd->product_desc);

	return 0;

 out:
	crisv32_hcd_remove(dev);
	return retval;
}

/* cleanup after the host controller and driver */
static int __init_or_module crisv32_hcd_remove(struct device *dev)
{
	struct crisv32_hcd *crisv32_hcd = dev_get_drvdata(dev);
	struct usb_hcd *hcd;

	if (!crisv32_hcd)
		return 0;
	hcd = crisv32_hcd_to_hcd(crisv32_hcd);

	usb_remove_hcd(hcd);

	if (crisv32_hcd->iop_init_done)
		iop_usb_exit();

	if (crisv32_hcd->descr_cache)
		kmem_cache_destroy(crisv32_hcd->descr_cache);

	usb_put_hcd(hcd);
	return 0;
}
 
/* register driver */
static int __init crisv32_hcd_init(void) 
{
	if (usb_disabled())
		return -ENODEV;

	printk(KERN_INFO "%s version "VERSION" "COPYRIGHT"\n", product_desc);

	crisv32_hc_platform_device =
		platform_device_register_simple((char *) hc_name, 0, NULL, 0);
	if (IS_ERR(crisv32_hc_platform_device))
		return PTR_ERR(crisv32_hc_platform_device);
	return driver_register(&crisv32_hc_device_driver);
	/* 
	 * Note that we do not set the DMA mask for the device,
	 * i.e. we pretend that we will use PIO, since no specific
	 * allocation routines are needed for DMA buffers. This will
	 * cause the HCD buffer allocation routines to fall back to
	 * kmalloc().
	 */
}

/* unregister driver */
static void __exit crisv32_hcd_exit(void) 
{	
	driver_unregister(&crisv32_hc_device_driver);
}
